import os
from abc import abstractmethod
from dataclasses import dataclass
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, TypeVar

import wandb
from git_t5.utils import rank_zero_only
from wandb.sdk.wandb_run import Run

if TYPE_CHECKING:
    from .trainer import T5Trainer
else:
    T5Trainer = Any

T = TypeVar("T", bound=Callable)


def rank_zero_experiment(fn: T) -> T:
    """Returns the real experiment on rank 0 and otherwise the DummyExperiment."""

    @wraps(fn)
    def experiment(self):
        @rank_zero_only
        def get_experiment():
            return fn(self)

        return get_experiment() or DummyExperiment()

    return experiment  # type: ignore


class DummyExperiment:
    """Dummy experiment"""

    def nop(self, *args: Any, **kwargs: Any) -> None:
        pass

    def __getattr__(self, _: Any) -> Any:
        return self.nop

    def __getitem__(self, idx: int) -> "DummyExperiment":
        # enables self.logger.experiment[0].add_image(...)
        return self


class BaseLogger:
    trainer: Optional[T5Trainer] = None

    @abstractmethod
    def log_metrics(
        self, metrics: Dict[str, float], step: Optional[int] = None
    ) -> None:
        raise NotImplementedError()

    def _add_prefix(
        self, metrics: Dict[str, float], prefix: Optional[str]
    ) -> Dict[str, float]:
        if prefix is not None:
            metrics = {f"{prefix}_{k}": v for k, v in metrics.items()}
        return metrics


@dataclass
class LoggerConfig:
    pass


@dataclass
class WandbLoggerConfig(LoggerConfig):
    name: Optional[str] = None
    save_dir: Optional[str] = None
    offline: Optional[bool] = None
    run_id: Optional[str] = None
    anonymous: Optional[bool] = None
    version: Optional[str] = None
    project: Optional[str] = None
    prefix: Optional[str] = None


class WandbLogger(BaseLogger):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        config: WandbLoggerConfig,
        experiment: Optional[Run] = None,
        **kwargs,
    ) -> None:
        anonymous_map = {
            True: "allow",
            False: None,
            None: None,
        }

        self.config = config
        self._wandb_init = dict(
            name=config.name,
            project=config.project,
            id=config.version or config.run_id,
            dir=config.save_dir,
            resume="allow",
            anonymous=anonymous_map.get(config.anonymous),
        )

        self._wandb_init.update(**kwargs)
        self._experiment = experiment

    @property
    @rank_zero_experiment
    def experiment(self) -> Run:
        if self._experiment is None:
            if self.config.offline:
                os.environ["WANDB_MODE"] = "dryrun"

            if wandb.run is None:
                self._experiment = wandb.init(**self._wandb_init)
            else:
                self._experiment = wandb.run

        self._experiment.define_metric("trainer/global_step")
        self._experiment.define_metric(
            "*",
            step_metric="trainer/global_step",
            step_sync=True,
        )

        return self._experiment

    @rank_zero_only
    def log_metrics(
        self, metrics: Dict[str, float], step: Optional[int] = None
    ) -> None:
        metrics = self._add_prefix(metrics, prefix=self.config.prefix)
        if step is not None:
            self.experiment.log({**metrics, "trainer/global_step": step})
        else:
            self.experiment.log(metrics)


class DummyLogger(BaseLogger):
    @rank_zero_only
    def log_metrics(
        self, metrics: Dict[str, float], step: Optional[int] = None
    ) -> None:
        print(metrics)



import json
import os
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import optax
from flax import jax_utils
from flax.serialization import from_bytes, to_bytes
from flax.training.common_utils import get_metrics
from flax.training.train_state import TrainState
from git_t5.core import AutoOptimizer, AutoScheduler
from git_t5.utils import rank_zero_only
from tqdm import tqdm

if TYPE_CHECKING:
    from git_t5.cli.train_model import Config
    from git_t5.core.data_module import T5DataModule
    from git_t5.core.logger import BaseLogger
    from git_t5.core.model import T5ModelForPreTraining
else:
    Config = Any
    T5DataModule = Any
    T5ModelForPreTraining = Any
    BaseLogger = Any


@dataclass
class TrainerConfig:
    pass


@dataclass
class T5TrainerConfig(TrainerConfig):
    max_epochs: int = 3
    logging_steps: int = 500
    save_steps: int = 500
    valid_steps: int = 2000


@dataclass
class T5Trainer:
    config: Config
    model: T5ModelForPreTraining
    data_module: T5DataModule
    logger: BaseLogger
    current_epoch: int = 0
    global_step: int = 0

    def __post_init__(self) -> None:
        self.model.trainer = self
        self.data_module.trainer = self
        self.logger.trainer = self

    def fit(self) -> None:
        num_epochs = self.config.trainer.max_epochs
        train_dataloader = self.data_module.train_dataloader()
        train_samples = len(train_dataloader)

        optimizer, scheduler_fn = self.configure_optimizers()
        state = self.create_state(optimizer, self.config.training.checkpoint_dir)
        rng = jax.random.PRNGKey(self.config.training.seed)
        dropout_rng = jax.random.split(rng, jax.device_count())

        train_time = 0
        train_start_time = time.time()
        train_metrics = []

        for epoch in range(num_epochs):
            self.current_epoch = epoch

            for batch in tqdm(
                train_dataloader,
                total=train_samples,
                desc=f"Training epoch={epoch}",
            ):
                state, metrics, dropout_rng = self.model.train_step(
                    state, dropout_rng, batch
                )

                # accumulate train metrics
                train_metrics.append(metrics)

                # increase the global train step
                self.global_step += 1

                if self.global_step % self.config.trainer.logging_steps == 0:
                    train_metrics = jax_utils.unreplicate(train_metrics)
                    state_step = jax_utils.unreplicate(state.step)
                    train_lr = scheduler_fn(state_step - 1)
                    train_time = time.time() - train_start_time

                    common_metrics = {
                        "epoch": epoch,
                        "step": state_step,
                        "train_lr": train_lr,
                        "train_time": train_time,
                    }

                    # write train metrics
                    self.log_metrics(
                        train_metrics,
                        step=self.global_step,
                        prefix="train",
                    )

                    # write common metrics
                    self.log_metrics(
                        common_metrics,
                        step=self.global_step,
                        prefix=None,
                    )

                    # clear train metrics buffer
                    train_metrics = []

                if self.global_step % self.config.trainer.valid_steps == 0:
                    self.validate(state)

                if self.global_step % self.config.trainer.save_steps == 0:
                    self.save_checkpoint(state, self.config.training.output_dir)

    def validate(self, state: TrainState) -> None:
        valid_dataloader = self.data_module.valid_dataloader()
        valid_samples = len(valid_dataloader)
        valid_metrics = []

        for batch in tqdm(
            valid_dataloader,
            total=valid_samples,
            desc="Validating...",
        ):
            metrics = self.model.valid_step(state, batch)
            valid_metrics.append(metrics)

        valid_metrics = get_metrics(valid_metrics)
        valid_metrics = jax.tree_map(jnp.mean, valid_metrics)

        self.log_metrics(
            valid_metrics,
            step=self.global_step,
            prefix="valid",
        )

    def configure_optimizers(
        self,
    ) -> Tuple[optax.GradientTransformation, optax.Schedule]:
        def training_steps() -> int:
            total_steps = len(self.data_module.datasets["train"])
            batch_size = self.data_module.config.data.train_batch_size
            total_batch_size = batch_size * jax.device_count()
            num_epochs = self.config.trainer.max_epochs
            num_train_steps = (total_steps // total_batch_size) * num_epochs
            return num_train_steps

        cfg = self.config.optimizer
        cfg.scheduler.train_steps = training_steps()
        optimizer = AutoOptimizer.from_config(cfg)
        scheduler = AutoScheduler.from_config(cfg.scheduler)

        return optimizer, scheduler

    @rank_zero_only
    def save_checkpoint(self, state: TrainState, save_dir: str) -> None:
        # save pretrained flax model
        state = jax_utils.unreplicate(state)
        flax_model = self.model.model
        flax_model.save_pretrained(
            save_dir,
            params=state.params,
            push_to_hub=self.config.training.push_to_hub,
            commit_message=f"Saving weights and logs of step {state.step}",
        )

        # unwrap state step scalar
        state_step = state.step
        if isinstance(state_step, jnp.ndarray):
            state_step = state_step.item()

        # save optimizer weights
        with open(os.path.join(save_dir, "opt_state.msgpack"), "wb") as f:
            f.write(to_bytes(state.opt_state))  # type: ignore

        # save the training state
        with open(os.path.join(save_dir, "training_state.json"), "w") as f:
            training_state = {
                "step": state_step,
                "global_step": self.global_step,
                "current_epoch": self.current_epoch,
            }

            json.dump(training_state, f)

    def restore_checkpoint(
        self, state: TrainState, save_dir: str
    ) -> Tuple[Dict[str, Any], optax.OptState, Dict[str, Any]]:
        # restore pretrained flax model weights
        with open(os.path.join(save_dir, "flax_model.msgpack"), "rb") as f:
            params = from_bytes(state.params, f.read())
            assert isinstance(params, dict)

        # restore optimizer weights
        with open(os.path.join(save_dir, "opt_state.msgpack"), "rb") as f:
            opt_state = from_bytes(state.opt_state, f.read())
            assert isinstance(opt_state, type(state.opt_state))

        # restore the training state
        with open(os.path.join(save_dir, "training_state.json"), "r") as f:
            training_state = json.load(f)
            assert isinstance(training_state, dict)

        return params, opt_state, training_state

    def create_state(
        self,
        optimizer: optax.GradientTransformation,
        checkpoint_dir: Optional[str] = None,
    ) -> TrainState:
        model = self.model.model
        state = TrainState.create(
            apply_fn=model.__call__,
            params=model.params,
            tx=optimizer,
        )

        if checkpoint_dir is not None:
            params, opt_state, training_state = self.restore_checkpoint(
                state, checkpoint_dir
            )

            step = training_state["step"]
            self.current_epoch = training_state["current_epoch"]
            self.global_step = training_state["global_step"]

            state = TrainState(
                step=step,
                apply_fn=model.__call__,
                params=params,
                tx=optimizer,
                opt_state=opt_state,
            )

        state = jax_utils.replicate(state)
        return state

    def log_metrics(
        self,
        metrics: Union[List[Dict[str, float]], Dict[str, float]],
        step: int,
        prefix: Optional[str] = None,
    ) -> None:
        device_metrics = metrics
        if not isinstance(device_metrics, list):
            device_metrics = [device_metrics]

        for idx, scalar_metrics in enumerate(device_metrics):
            current_step = step - len(device_metrics) + idx + 1
            scalar_metrics = self.prepare_metrics(scalar_metrics, prefix)
            self.logger.log_metrics(scalar_metrics, step=current_step)

    def prepare_metrics(
        self,
        metrics: Dict[str, float],
        prefix: Optional[str] = None,
    ) -> Dict[str, float]:
        prefix = prefix + "_" if prefix else ""
        return {f"{prefix}{name}": value for name, value in metrics.items()}



import json
from dataclasses import dataclass, field
from typing import Dict, Iterator, List, Optional, Union

from tokenizers import (
    AddedToken,
    Regex,
    Tokenizer,
    decoders,
    normalizers,
    pre_tokenizers,
    processors,
    trainers,
)
from tokenizers.implementations.base_tokenizer import BaseTokenizer
from tokenizers.models import BPE
from tokenizers.processors import TemplateProcessing
from typing_extensions import TypedDict

Token = Union[str, AddedToken]


NORMALIZERS = {
    "nfc": normalizers.NFC,
    "nfd": normalizers.NFD,
    "nfkc": normalizers.NFKC,
    "nfkd": normalizers.NFKD,
}


def unicode_normalizer(normalizer: str) -> normalizers.Normalizer:
    if normalizer not in NORMALIZERS:
        raise ValueError(
            "{} is not a known unicode normalizer. Available are {}".format(
                normalizer, NORMALIZERS.keys()
            )
        )

    return NORMALIZERS[normalizer]()


class SpecialToken(TypedDict):
    id: int
    token: str


@dataclass
class TokenizerConfig:
    vocab_size: int = 32_000
    special_tokens: List[str] = field(default_factory=list)
    additional_special_tokens: List[str] = field(default_factory=list)
    model_max_length: Optional[int] = None
    show_progress: bool = True


@dataclass
class SentencePieceTokenizerConfig(TokenizerConfig):
    dropout: Optional[float] = None
    add_prefix_space: bool = False
    trim_offsets: bool = False
    min_frequency: int = 2
    lowercase: bool = False
    remove_extra_spaces: bool = True
    unicode_normalizer: Optional[str] = "nfkc"
    unk_token: str = "<unk>"
    eos_token: str = "</s>"
    pad_token: str = "<pad>"


# pylint: disable=too-many-arguments
class SentencePieceTokenizer(BaseTokenizer):
    def __init__(self, config: SentencePieceTokenizerConfig) -> None:
        self.config = config

        # sentencepiece special tokens map
        self.special_tokens_map: Dict[str, SpecialToken] = {
            "pad": {"id": 0, "token": config.pad_token},
            "eos": {"id": 1, "token": config.eos_token},
            "unk": {"id": 2, "token": config.unk_token},
        }

        # sentencepiece special tokens list
        self.special_tokens = self.get_special_tokens()
        self.special_tokens += list(config.special_tokens)

        # sentencepiece bpe tokenizer
        tokenizer = self.get_tokenizer()

        parameters = {
            "model": "ByteLevelBPE",
            "add_prefix_space": config.add_prefix_space,
            "dropout": config.dropout,
            "trim_offsets": config.trim_offsets,
        }

        super().__init__(tokenizer, parameters)

    def get_special_tokens(self) -> List[Token]:
        special_tokens: List[Token] = [""] * len(self.special_tokens_map)
        for token_dict in self.special_tokens_map.values():
            special_tokens[token_dict["id"]] = token_dict["token"]
        return special_tokens

    def get_tokenizer(self) -> Tokenizer:
        # configure the sentencepiece bpe tokenizer
        tokenizer = Tokenizer(
            BPE(
                dropout=self.config.dropout,
                unk_token=self.config.unk_token,
            )
        )

        tokenizer.normalizer = self.get_normalizer()  # type: ignore
        tokenizer.pre_tokenizer = self.get_pre_tokenizer()  # type: ignore
        tokenizer.decoder = self.get_decoder()  # type: ignore
        tokenizer.post_processor = self.get_post_processor()  # type: ignore

        return tokenizer

    def get_normalizer(self) -> normalizers.Normalizer:
        normalizer_list: List[normalizers.Normalizer] = []
        normalizer_list.append(normalizers.Nmt())

        if self.config.unicode_normalizer is not None:
            normalizer_list.append(unicode_normalizer(self.config.unicode_normalizer))

        if self.config.remove_extra_spaces:
            normalizer_list.append(normalizers.Replace(Regex(" {2,}"), " "))

        if self.config.lowercase:
            normalizer_list.append(normalizers.Lowercase())

        return normalizers.Sequence(normalizer_list)  # type: ignore

    def get_pre_tokenizer(self) -> pre_tokenizers.PreTokenizer:
        return pre_tokenizers.ByteLevel()

    def get_decoder(self) -> decoders.Decoder:
        return decoders.ByteLevel()

    def get_post_processor(self) -> processors.PostProcessor:
        return TemplateProcessing(  # type: ignore
            single=f"$A {self.special_tokens_map['eos']['token']}",
            special_tokens=[
                (
                    self.special_tokens_map["eos"]["token"],
                    self.special_tokens_map["eos"]["id"],
                )
            ],
        )

    def train_from_files(self, files: Union[str, List[str]]) -> None:
        """Train the model using the given files"""

        trainer = trainers.BpeTrainer(
            vocab_size=self.config.vocab_size,
            min_frequency=self.config.min_frequency,
            show_progress=self.config.show_progress,
            special_tokens=self.special_tokens,
            initial_alphabet=pre_tokenizers.ByteLevel.alphabet(),
        )

        if isinstance(files, str):
            files = [files]

        self._tokenizer.train(files, trainer=trainer)
        self._add_token_id("unk")

    def train_from_iterator(
        self, iterator: Union[Iterator[str], Iterator[Iterator[str]]]
    ) -> None:
        """Train the model using the given iterator"""

        trainer = trainers.BpeTrainer(
            vocab_size=self.config.vocab_size,
            min_frequency=self.config.min_frequency,
            show_progress=self.config.show_progress,
            special_tokens=self.special_tokens,
            initial_alphabet=pre_tokenizers.ByteLevel.alphabet(),
        )

        self._tokenizer.train_from_iterator(iterator, trainer=trainer)
        self._add_token_id("unk")

    def _add_token_id(self, token: str) -> None:
        token_id = f"{token}_id"
        tokenizer_json = json.loads(self._tokenizer.to_str())
        tokenizer_json["model"][token_id] = self.special_tokens_map[token]["id"]
        self._tokenizer = Tokenizer.from_str(json.dumps(tokenizer_json))



from dataclasses import dataclass
from typing import Optional

import chex
import jax.numpy as jnp
import optax
from git_t5.utils import resolve_object


def inverse_square_root_schedule(
    init_value: chex.Scalar,
    end_value: chex.Scalar,
    transition_steps: int,
    transition_begin: int = 0,
) -> optax.Schedule:
    """Constructs a schedule with inverse square root transition from init to end value.
    Args:
      init_value: initial value for the scalar to be annealed.
      end_value: end value of the scalar to be annealed.
      transition_steps: number of steps over which annealing takes place,
        the scalar starts changing at `transition_begin` steps and completes
        the transition by `transition_begin + transition_steps` steps.
        If `transition_steps <= 0`, then the entire annealing process is disabled
        and the value is held fixed at `init_value`.
      transition_begin: must be positive. After how many steps to start annealing
        (before this many steps the scalar value is held fixed at `init_value`).
    Returns:
      schedule: A function that maps step counts to values.
    """

    if transition_steps <= 0:
        return lambda _: init_value

    if transition_begin < 0:
        transition_begin = 0

    decay_factor = (init_value - end_value) * transition_begin ** 0.5

    def schedule(count: chex.Numeric) -> chex.Numeric:
        return decay_factor * 1 / jnp.sqrt(count + transition_begin)

    return schedule


@dataclass
class SchedulerConfig:
    train_steps: Optional[int] = None
    learning_rate: Optional[float] = None


@dataclass
class PolynomialSchedulerConfig(SchedulerConfig):
    init_learning_rate: float = 0.0
    warmup_steps: int = 0
    power: float = 1.0


@dataclass
class LinearSchedulerConfig(SchedulerConfig):
    init_learning_rate: float = 0.0
    warmup_steps: int = 0


@dataclass
class InverseSquareRootSchedulerConfig(SchedulerConfig):
    warmup_steps: int = 0


@dataclass
class ConstantSchedulerConfig(SchedulerConfig):
    init_learning_rate: float = 0.0
    warmup_steps: int = 0


def polynomial_scheduler(config: PolynomialSchedulerConfig) -> optax.Schedule:
    if config.train_steps is None:
        raise ValueError("`train_steps` must be specified.")
    if config.learning_rate is None:
        raise ValueError("`learning_rate` must be specified.")

    warmup_fn = optax.linear_schedule(
        init_value=config.init_learning_rate,
        end_value=config.learning_rate,
        transition_steps=config.warmup_steps,
    )
    decay_fn = optax.polynomial_schedule(
        init_value=config.learning_rate,
        end_value=config.init_learning_rate,
        power=config.power,
        transition_steps=config.train_steps - config.warmup_steps,
    )
    scheduler_fn = optax.join_schedules(
        schedules=[warmup_fn, decay_fn],
        boundaries=[config.warmup_steps],
    )

    return scheduler_fn


def linear_scheduler(config: LinearSchedulerConfig) -> optax.Schedule:
    if config.train_steps is None:
        raise ValueError("`train_steps` must be specified.")
    if config.learning_rate is None:
        raise ValueError("`learning_rate` must be specified.")

    warmup_fn = optax.linear_schedule(
        init_value=config.init_learning_rate,
        end_value=config.learning_rate,
        transition_steps=config.warmup_steps,
    )
    decay_fn = optax.linear_schedule(
        init_value=config.learning_rate,
        end_value=config.init_learning_rate,
        transition_steps=config.train_steps - config.warmup_steps,
    )
    scheduler_fn = optax.join_schedules(
        schedules=[warmup_fn, decay_fn],
        boundaries=[config.warmup_steps],
    )

    return scheduler_fn


def inverse_square_root_scheduler(
    config: InverseSquareRootSchedulerConfig,
) -> optax.Schedule:
    if config.train_steps is None:
        raise ValueError("`train_steps` must be specified.")
    if config.learning_rate is None:
        raise ValueError("`learning_rate` must be specified.")

    warmup_fn = optax.linear_schedule(
        init_value=config.learning_rate,
        end_value=config.learning_rate,
        transition_steps=config.warmup_steps,
    )
    decay_fn = inverse_square_root_schedule(
        init_value=config.learning_rate,
        end_value=0.0,
        transition_steps=config.train_steps,
        transition_begin=config.warmup_steps,
    )
    scheduler_fn = optax.join_schedules(
        schedules=[warmup_fn, decay_fn],
        boundaries=[config.warmup_steps],
    )

    return scheduler_fn


def constant_scheduler(config: ConstantSchedulerConfig) -> optax.Schedule:
    if config.train_steps is None:
        raise ValueError("`train_steps` must be specified.")
    if config.learning_rate is None:
        raise ValueError("`learning_rate` must be specified.")

    warmup_fn = optax.linear_schedule(
        init_value=config.init_learning_rate,
        end_value=config.learning_rate,
        transition_steps=config.warmup_steps,
    )
    decay_fn = optax.linear_schedule(
        init_value=config.learning_rate,
        end_value=config.learning_rate,
        transition_steps=config.train_steps - config.warmup_steps,
    )
    scheduler_fn = optax.join_schedules(
        schedules=[warmup_fn, decay_fn],
        boundaries=[config.warmup_steps],
    )

    return scheduler_fn


class AutoScheduler:
    @staticmethod
    def from_config(config: SchedulerConfig) -> optax.Schedule:
        config = resolve_object(config)
        scheduler: optax.Schedule
        if isinstance(config, PolynomialSchedulerConfig):
            scheduler = polynomial_scheduler(config)
        elif isinstance(config, LinearSchedulerConfig):
            scheduler = linear_scheduler(config)
        elif isinstance(config, InverseSquareRootSchedulerConfig):
            scheduler = inverse_square_root_scheduler(config)
        elif isinstance(config, ConstantSchedulerConfig):
            scheduler = constant_scheduler(config)
        else:
            raise ValueError(f"Unknown scheduler type: {type(config)}.")
        return scheduler



from collections import defaultdict
from typing import Dict, List, TypeVar

K = TypeVar("K")
T = TypeVar("T")


def stack_mappings(mappings: List[Dict[K, T]]) -> Dict[K, List[T]]:
    def flatten(sequence: List[List[T]]) -> List[T]:
        return [item for subsequence in sequence for item in subsequence]

    result = defaultdict(list)
    keys = set(flatten([mapping.keys() for mapping in mappings]))  # type: ignore
    for mapping in mappings:
        for key in keys:
            result[key].append(mapping.get(key, None))

    return dict(result)



import os
from dataclasses import dataclass

import hydra
from git_t5.core import (
    SentencePieceTokenizerConfig,
    SentencePieceTrainer,
    SentencePieceTrainerConfig,
)
from hydra.core.config_store import ConfigStore

from .config import DefaultConfig, register_base_configs


@dataclass
class Config(DefaultConfig):
    tokenizer_trainer: SentencePieceTrainerConfig = SentencePieceTrainerConfig()


def register_configs() -> None:
    cs = ConfigStore.instance()
    cs.store(name="default", node=Config)
    cs.store(
        group="tokenizer_trainer",
        name="base_tokenizer_trainer",
        node=SentencePieceTrainerConfig,
    )
    cs.store(
        group="tokenizer_trainer/tokenizer",
        name="base_tokenizer",
        node=SentencePieceTokenizerConfig,
    )


@hydra.main(config_path="conf", config_name="config_tokenizer")
def hydra_entry(cfg: Config) -> None:
    trainer = SentencePieceTrainer.from_config(cfg.tokenizer_trainer)
    trainer.train()


def main() -> None:
    # enable rust iterators multithreading
    os.environ["TOKENIZERS_PARALLELISM"] = "true"

    register_base_configs()
    register_configs()
    hydra_entry()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    main()



from typing import TypeVar

from omegaconf import DictConfig

T = TypeVar("T")


def resolve_object(obj: T) -> T:
    if isinstance(obj, DictConfig):
        obj = obj._to_object()  # pylint: disable=protected-access
    return obj



import copy
import typing
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

import datasets as hfds
import jax
import numpy as np
from git_t5.core import HFDatasetConfig, LocalDatasetConfig, MultitaskDatasetConfig
from git_t5.data import (
    DataCollatorForT5MLM,
    Dataset,
    T5Dataset,
    T5MultitaskDataset,
    compute_input_and_target_lengths,
    prepare_dataset,
)
from git_t5.utils import resolve_object, stack_mappings
from omegaconf import MISSING
from torch.utils import data
from transformers import AutoTokenizer, PreTrainedTokenizerBase
from transformers.tokenization_utils_base import VERY_LARGE_INTEGER

if TYPE_CHECKING:
    from git_t5.cli.train_model import Config
    from git_t5.core.trainer import T5Trainer
else:
    Config = Any
    T5Trainer = Any

T = typing.TypeVar("T")


def tokenize_fn(
    tokenizer: PreTrainedTokenizerBase,
    column: str,
) -> Callable[..., Dict[str, Union[List[List[int]], np.ndarray]]]:
    def wrap_fn(
        examples: Dict[str, List[str]]
    ) -> Dict[str, Union[List[List[int]], np.ndarray]]:
        return tokenizer(
            examples[column],
            truncation=False,
            return_attention_mask=False,
        )  # type: ignore

    return wrap_fn


def select_subset(
    dataset: hfds.Dataset,
    size: Union[float, int],
    seed: Optional[int] = None,
) -> hfds.Dataset:
    num_samples: int
    if isinstance(size, int) or size > 1:
        num_samples = min(int(size), len(dataset))
    else:
        num_samples = int(len(dataset) * size)

    rng = np.random.default_rng(seed)
    indices = rng.integers(0, len(dataset), (num_samples,))

    return dataset.select(indices)


def _collate_fn(samples: List[T]) -> T:
    assert len(samples) == 1
    return samples[0]


@dataclass
class DataModuleConfig:
    pass


@dataclass
class T5DataModuleConfig(DataModuleConfig):
    validation_size: float = 0.05
    max_sequence_length: Optional[int] = None
    train_batch_size: int = 8
    valid_batch_size: int = 8
    num_proc: Optional[int] = None
    num_workers: Optional[int] = None
    mlm_probability: float = 0.15
    mean_noise_span_length: float = 3.0
    decoder_start_token_id: int = MISSING
    limit_train_size: float = 1.0
    limit_valid_size: float = 1.0


@dataclass
class T5DataModule:
    config: Config
    datasets: Dict[str, Dataset]
    tokenizer: PreTrainedTokenizerBase
    data_collator: DataCollatorForT5MLM
    max_sequence_length: int
    input_length: int
    target_length: int
    trainer: Optional[T5Trainer] = None

    @classmethod
    def from_config(cls, config: Config) -> "T5DataModule":
        tokenizer = cls.load_tokenizer(config)
        max_sequence_length = config.data.max_sequence_length or VERY_LARGE_INTEGER
        max_sequence_length = min(max_sequence_length, tokenizer.model_max_length)

        input_length, target_length = compute_input_and_target_lengths(
            max_sequence_length,
            noise_density=config.data.mlm_probability,
            mean_noise_span_length=config.data.mean_noise_span_length,
            extra_tokens_per_span_inputs=1,
            extra_tokens_per_span_targets=1,
        )

        eos_token_id = tokenizer.eos_token_id
        pad_token_id = tokenizer.pad_token_id
        sentinel_token_id = tokenizer.convert_tokens_to_ids("<extra_id_0>")  # type: ignore
        if eos_token_id is None:
            raise ValueError("Tokenizer must have an existing `eos_token_id` value.")
        if pad_token_id is None:
            raise ValueError("Tokenizer must have an existing `pad_token_id` value.")
        if sentinel_token_id is None:
            raise ValueError("Tokenizer must have an existing `eos_token_id` value.")

        data_collator = DataCollatorForT5MLM(
            tokenizer=tokenizer,
            noise_density=config.data.mlm_probability,
            mean_noise_span_length=config.data.mean_noise_span_length,
            input_length=max_sequence_length,
            target_length=target_length,
            eos_token_id=eos_token_id,
            pad_token_id=pad_token_id,
            sentinel_token_id=sentinel_token_id,
            decoder_start_token_id=config.data.decoder_start_token_id,
        )

        datasets = cls.load_datasets(config, tokenizer, input_length)
        datasets = stack_mappings(datasets)
        datasets = {
            "train": cls.get_dataset(
                datasets["train"],
                batch_size=config.data.train_batch_size,
                collate_fn=data_collator,
                shuffle=True,
                drop_last=True,
                seed=config.training.seed,
            ),
            "valid": cls.get_dataset(
                datasets["validation"],
                batch_size=config.data.valid_batch_size,
                collate_fn=data_collator,
                shuffle=False,
                drop_last=True,
                seed=None,
            ),
        }

        return T5DataModule(
            config,
            datasets=datasets,
            tokenizer=tokenizer,
            data_collator=data_collator,
            max_sequence_length=max_sequence_length,
            input_length=input_length,
            target_length=target_length,
        )

    def train_dataloader(self) -> data.DataLoader:
        num_workers = self.config.data.num_workers or 0
        return data.DataLoader(
            self.datasets["train"],
            collate_fn=_collate_fn,
            num_workers=num_workers,
        )

    def valid_dataloader(self) -> data.DataLoader:
        num_workers = self.config.data.num_workers or 0
        return data.DataLoader(
            self.datasets["valid"],
            collate_fn=_collate_fn,
            num_workers=num_workers,
        )

    @classmethod
    def get_dataset(
        cls,
        datasets: List[hfds.Dataset],
        batch_size: int,
        collate_fn: typing.Callable[..., Dict[str, np.ndarray]],
        shuffle: bool,
        drop_last: bool,
        seed: Optional[int],
    ) -> Dataset:
        batch_size = batch_size * jax.device_count()

        if len(datasets) == 1:
            dataset = T5Dataset(
                datasets[0],
                batch_size=batch_size,
                collate_fn=collate_fn,
                shuffle=shuffle,
                drop_last=drop_last,
                seed=seed,
            )
        else:
            dataset = T5MultitaskDataset(
                datasets,
                batch_size=batch_size,
                collate_fn=collate_fn,
                shuffle=shuffle,
                drop_last=drop_last,
                seed=seed,
            )

        return dataset

    @classmethod
    def prepare_dataset(
        cls,
        dataset: hfds.DatasetDict,
        config: Config,
        tokenizer: PreTrainedTokenizerBase,
        input_length: int,
    ) -> hfds.DatasetDict:
        if config.dataset.column_name is None:
            raise ValueError(
                "You must provide a `column_name` to specify which column of the dataset to use."
            )

        dataset = prepare_dataset(
            dataset,
            tokenize_fn(tokenizer, config.dataset.column_name),
            input_length=input_length,
            batch_size=128,
            load_from_cache_file=not config.training.overwrite_cache,
            num_proc=config.data.num_proc,
        )

        # limit preprocessed dataset if needed
        dataset = cls.limit_dataset(dataset, config)

        return dataset

    @classmethod
    def load_datasets(
        cls,
        config: Config,
        tokenizer: PreTrainedTokenizerBase,
        input_length: int,
    ) -> List[Dict[str, hfds.Dataset]]:
        dataset_config = resolve_object(config.dataset)
        datasets: List[Dict[str, hfds.Dataset]] = []
        if isinstance(dataset_config, MultitaskDatasetConfig):
            dataset_config.resolve()
            for _, dataset_config in dataset_config.tasks.items():
                config = copy.deepcopy(config)
                config.dataset = dataset_config
                dataset = cls.load_dataset(config)
                dataset = cls.prepare_dataset(dataset, config, tokenizer, input_length)
                datasets.append(dataset)
        else:
            dataset = cls.load_dataset(config)
            dataset = cls.prepare_dataset(dataset, config, tokenizer, input_length)
            datasets.append(dataset)

        return datasets

    @classmethod
    def load_dataset(cls, config: Config) -> hfds.DatasetDict:
        dataset_config = resolve_object(config.dataset)
        if isinstance(dataset_config, LocalDatasetConfig):
            dataset = hfds.load_from_disk(dataset_config.dataset_path)
            if not isinstance(dataset, hfds.DatasetDict):
                dataset = hfds.DatasetDict(train=dataset)

            if "validation" not in dataset.keys():
                dataset = dataset["train"].train_test_split(
                    test_size=config.data.validation_size,
                    load_from_cache_file=not config.training.overwrite_cache,
                )
                dataset["validation"] = dataset.pop("test")
        elif isinstance(dataset_config, HFDatasetConfig):
            dataset = hfds.load_dataset(
                dataset_config.dataset_name,
                dataset_config.dataset_config,
                cache_dir=config.training.cache_dir,
            )

            if not isinstance(dataset, hfds.DatasetDict):
                dataset = hfds.DatasetDict(train=dataset)

            if "validation" not in dataset.keys():
                valid_percentage = int(config.data.validation_size * 100)
                dataset["validation"] = hfds.load_dataset(
                    dataset_config.dataset_name,
                    dataset_config.dataset_config,
                    split=f"train[:{valid_percentage}%]",
                    cache_dir=config.training.cache_dir,
                )
                dataset["train"] = hfds.load_dataset(
                    dataset_config.dataset_name,
                    dataset_config.dataset_config,
                    split=f"train[{valid_percentage}%:]",
                    cache_dir=config.training.cache_dir,
                )
        else:
            raise ValueError("Unknown dataset type provided.")

        # limit loaded dataset if needed
        dataset = cls.limit_dataset(dataset, config)

        return dataset

    @classmethod
    def limit_dataset(
        cls,
        dataset: hfds.DatasetDict,
        config: Config,
    ) -> hfds.DatasetDict:
        if config.data.limit_train_size != 1:
            dataset["train"] = select_subset(
                dataset["train"],
                config.data.limit_train_size,
                config.training.seed,
            )

        if config.data.limit_valid_size != 1:
            dataset["validation"] = select_subset(
                dataset["validation"],
                config.data.limit_valid_size,
                config.training.seed,
            )

        return dataset

    @classmethod
    def load_tokenizer(cls, config: Config) -> PreTrainedTokenizerBase:
        tokenizer = AutoTokenizer.from_pretrained(
            config.tokenizer.tokenizer_path,
            use_fast=config.tokenizer.use_fast,
            cache_dir=config.training.cache_dir,
        )

        return tokenizer



import math
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union

import datasets
import numpy as np

Dataset = TypeVar("Dataset", bound=Union[datasets.Dataset, datasets.DatasetDict])


def column_names(dataset: Any) -> List[str]:
    if isinstance(dataset, datasets.DatasetDict):
        values = dataset["train"].column_names
    elif isinstance(dataset, datasets.Dataset):
        values = dataset.column_names
    else:
        raise ValueError("Unsupported dataset type.")
    return values


def pad_batch(
    sequences: List[Union[List[int], np.ndarray]],
    pad: int,
    pad_first: bool = False,
) -> np.ndarray:
    batch_size = len(sequences)
    max_length = max(len(sequence) for sequence in sequences)
    batch = np.full((batch_size, max_length), pad, dtype=np.int32)
    for idx, sequence in enumerate(sequences):
        sequence = sequence[:-1]
        length = len(sequence)
        if pad_first:
            batch[idx, -length:] = sequence
        else:
            batch[idx, :length] = sequence
    return batch


def select_random_chunk(
    dataset: Dataset,
    feature_key: str,
    max_sequence_length: int,
    seed: Optional[int] = None,
    uniform_random_start: bool = False,
    load_from_cache_file: bool = True,
    num_proc: Optional[int] = None,
) -> Dataset:
    if max_sequence_length is None:
        raise ValueError("Must specify `max_sequence_length`.")

    def map_fn(
        features: Dict[str, Union[List[int], np.ndarray]]
    ) -> Dict[str, np.ndarray]:
        tokens = np.asarray(features[feature_key])
        num_tokens = np.size(tokens)
        rng = np.random.default_rng(seed)
        if uniform_random_start:
            start = rng.integers(
                low=-max_sequence_length + 1, high=num_tokens, dtype=np.int32
            )
            end = np.minimum(start + max_sequence_length, num_tokens)
            start = np.maximum(start, 0)
        else:
            num_segments = np.ceil(np.divide(num_tokens, max_sequence_length))
            num_segments = num_segments.astype(np.int32)
            start = rng.integers(low=0, high=num_segments, dtype=np.int32)
            end = np.minimum(start + max_sequence_length, num_tokens)
        chunk = {feature_key: tokens[start:end]}
        return chunk

    dataset = dataset.filter(
        lambda x: np.not_equal(np.size(x[feature_key]), 0).item(),
        load_from_cache_file=load_from_cache_file,
        num_proc=num_proc,
    )
    dataset = dataset.map(
        map_fn,
        load_from_cache_file=load_from_cache_file,
        num_proc=num_proc,
    )

    return dataset


def split_tokens(
    dataset: Dataset,
    feature_key: str,
    batch_size: int,
    pad_token_id: int,
    min_tokens_per_segment: Optional[int],
    max_tokens_per_segment: int,
    seed: Optional[int] = None,
    drop_last: bool = True,
    load_from_cache_file: bool = True,
    num_proc: Optional[int] = None,
) -> Dataset:
    def map_fn(
        features: Dict[str, List[Union[List[int], np.ndarray]]]
    ) -> Dict[str, List[np.ndarray]]:
        # pad and roll the batch by the `feature_key`
        tokens = pad_batch(features[feature_key], pad_token_id)
        tokens = np.reshape(tokens, [-1])
        tokens = tokens[tokens != pad_token_id]
        num_tokens = np.size(tokens)

        # calculate the length of segments
        if min_tokens_per_segment is None:
            length: int = max_tokens_per_segment
        else:
            rng = np.random.default_rng(seed)
            length: int = (
                rng.uniform(
                    low=math.log(min_tokens_per_segment),
                    high=math.log(max_tokens_per_segment),
                    size=(1,),
                )
                .astype(np.int32)
                .item()
            )

        # split chunks into padded segments (batches)
        num_segments = np.ceil(np.divide(num_tokens, length))
        num_segments = num_segments.astype(np.int32)
        padding = num_segments * length - num_tokens
        padded = np.pad(tokens, [[0, padding]])

        outputs = np.reshape(padded, [-1, length])
        lengths = np.concatenate(
            [np.repeat(length, num_segments - 1), [length - padding]], axis=0
        )

        # unwrap the padded segments + drop the last rows if needed
        input_ids: List[np.ndarray] = []
        for output, length in zip(outputs, lengths):
            if drop_last and length != max_tokens_per_segment:
                continue

            input_ids.append(output[:length])

        return {feature_key: input_ids}

    dataset = dataset.map(
        map_fn,
        batched=True,
        batch_size=batch_size,
        remove_columns=column_names(dataset),
        load_from_cache_file=load_from_cache_file,
        num_proc=num_proc,
    )

    return dataset


def noise_span_to_unique_sentinel(
    tokens: np.ndarray,
    noise_mask: np.ndarray,
    sentinel_id: int,
    tokens_reversed: bool,
) -> np.ndarray:
    prev_token_is_noise = np.pad(noise_mask[:-1], [[1, 0]])
    first_noise_tokens = np.logical_and(noise_mask, np.logical_not(prev_token_is_noise))
    subsequent_noise_tokens = np.logical_and(noise_mask, prev_token_is_noise)
    if tokens_reversed:
        sentinel = sentinel_id + 1 - np.cumsum(first_noise_tokens)
    else:
        sentinel = sentinel_id - 1 + np.cumsum(first_noise_tokens)
    tokens = np.where(first_noise_tokens, sentinel, tokens)
    return tokens[~subsequent_noise_tokens]


def nonnoise_span_to_unique_sentinel(
    tokens: np.ndarray,
    noise_mask: np.ndarray,
    sentinel_id: int,
    tokens_reversed: bool,
) -> np.ndarray:
    return noise_span_to_unique_sentinel(
        tokens,
        np.logical_not(noise_mask),
        sentinel_id,
        tokens_reversed,
    )


def prepare_dataset(
    dataset: Dataset,
    tokenize_fn: Callable[..., Dict[str, Union[List[List[int]], np.ndarray]]],
    input_length: int,
    batch_size: int,
    drop_last: bool = True,
    load_from_cache_file: bool = True,
    num_proc: Optional[int] = None,
) -> Dataset:
    ds = dataset.map(
        tokenize_fn,
        batched=False,
        remove_columns=column_names(dataset),
        load_from_cache_file=load_from_cache_file,
        num_proc=num_proc,
    )
    ds = select_random_chunk(
        ds,
        feature_key="input_ids",
        max_sequence_length=65536,
        uniform_random_start=False,
        load_from_cache_file=load_from_cache_file,
        num_proc=num_proc,
    )
    ds = split_tokens(
        ds,
        feature_key="input_ids",
        batch_size=batch_size,
        pad_token_id=0,
        min_tokens_per_segment=None,
        max_tokens_per_segment=input_length,
        drop_last=drop_last,
        load_from_cache_file=load_from_cache_file,
        num_proc=num_proc,
    )
    return ds


def denoise(
    dataset: Dataset,
    noise_density: float,
    eos_token_id: int,
    noise_mask_fn: Callable[..., np.ndarray],
    inputs_fn: Callable[..., np.ndarray],
    targets_fn: Callable[..., np.ndarray],
    load_from_cache_file: bool = True,
    num_proc: Optional[int] = None,
) -> Dataset:
    def map_fn(
        features: Dict[str, Union[List[int], np.ndarray]]
    ) -> Dict[str, np.ndarray]:
        tokens = features["input_ids"]
        noise_mask = noise_mask_fn(np.size(tokens), noise_density)

        input_ids = inputs_fn(tokens, noise_mask)
        input_ids = np.concatenate([input_ids, [eos_token_id]], axis=0)

        if targets_fn is not None:
            labels = targets_fn(tokens, noise_mask)
        else:
            labels = tokens
        labels = np.concatenate((labels, [eos_token_id]), axis=0)

        return {"input_ids": input_ids, "labels": labels}

    dataset = dataset.map(
        map_fn,
        load_from_cache_file=load_from_cache_file,
        num_proc=num_proc,
    )

    return dataset


def compute_input_and_target_lengths(
    inputs_length: int,
    noise_density: float,
    mean_noise_span_length: float,
    extra_tokens_per_span_inputs: int,
    extra_tokens_per_span_targets: int,
) -> Tuple[int, int]:
    def compute_lengths(tokens_length: int) -> Tuple[int, int]:
        num_noise_tokens = int(round(tokens_length * noise_density))
        num_nonnoise_tokens = tokens_length - num_noise_tokens
        num_noise_spans = int(round(num_noise_tokens / mean_noise_span_length))
        # inputs contain all nonnoise tokens, sentinels for all noise spans
        # and one EOS token.
        return (
            num_nonnoise_tokens + num_noise_spans * extra_tokens_per_span_inputs + 1,
            num_noise_tokens + num_noise_spans * extra_tokens_per_span_targets + 1,
        )

    tokens_length = inputs_length
    while compute_lengths(tokens_length + 1)[0] <= inputs_length:
        tokens_length += 1

    inputs_length, targets_length = compute_lengths(tokens_length)

    # minor hack to get the targets length to be equal to inputs length
    # which is more likely to have been set to a nice round number.
    if noise_density == 0.5 and targets_length > inputs_length:
        tokens_length -= 1
        targets_length -= 1
    return tokens_length, targets_length



# Copyright (c) 2019-present, Facebook, Inc. and The FormerMagic Inc. team.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import re
import tokenize
from abc import abstractmethod
from enum import Enum
from io import BytesIO
from typing import Dict, List, Union

from sacrebleu.tokenizers import TokenizerV14International

PYTHON_CHAR2TOKEN = {
    "#": "<STOKEN0>",
    "\\n": "<STOKEN1>",
    '"""': "<STOKEN2>",
    "'''": "<STOKEN3>",
}

PYTHON_TOKEN2CHAR = {
    "<STOKEN0>": "#",
    "<STOKEN1>": "\\n",
    "<STOKEN2>": '"""',
    "<STOKEN3>": "'''",
}


class SpecialToken(Enum):
    INDENT = "<indent>"
    DEDENT = "<dedent>"
    NEWLINE = "<newline>"
    STR_NEWLINE = "<strnewline>"
    TAB_SYMBOL = "<tabsymbol>"
    SPACE = "▁"
    SPACE_TOKEN = "<spacetoken>"
    END_COM = "<endcom>"
    END_MARKER = "<endmarker>"


class Tokenizer:
    def __init__(
        self,
        special_tokens: List[str],
        char2token: Dict[str, str],
        token2char: Dict[str, str],
    ) -> None:
        self.special_tokens = special_tokens
        self.char2token = char2token
        self.token2char = token2char
        self.tokenizer_v14_international = TokenizerV14International()

    @abstractmethod
    def tokenize(
        self,
        text: str,
        keep_comments: bool,
        add_token_space: bool,
    ) -> str:
        raise NotImplementedError()

    @abstractmethod
    def detokenize(
        self,
        tokens: Union[List[str], str],
        add_token_space: bool,
    ) -> str:
        raise NotImplementedError()

    def stringify(self, tokens: List[str], add_token_space: bool) -> str:
        """Joins tokens into line with or without spaces aroung special tokens."""
        line = " ".join(tokens)
        if not add_token_space:
            tokens_pattern = "|".join(self.special_tokens)
            line = re.sub(r"\s*(%s)\s*" % tokens_pattern, r"\1", line)
        return line

    def destringify(self, line: str, add_token_space: bool) -> str:
        """Adds spaces to special tokens in the tokenized line if they are missing."""
        if not add_token_space:
            tokens_pattern = "|".join(self.special_tokens)
            line = re.sub(r"\s*(%s)\s*" % tokens_pattern, r" \1 ", line)
        return line

    def process_string(
        self,
        token: str,
        is_comment: bool,
        use_bleu_tokenization: bool = False,
    ) -> str:
        if is_comment:
            token = re.sub(" +", " ", token)
            token = re.sub(r"(.)\1\1\1\1+", r"\1\1\1\1\1", token)
            if len(re.sub(r"\W", "", token)) < 2:
                return ""

        token = token.replace(" ", f" {SpecialToken.SPACE.value} ")
        for char, special_token in self.char2token.items():
            token = token.replace(char, special_token)

        if token.startswith("<STOKEN0>"):
            if token.endswith("\n"):
                token = token[:-1]
            token += f" {SpecialToken.END_COM.value}"

        token = token.replace("\n", f" {SpecialToken.STR_NEWLINE.value} ")
        token = token.replace("\t", f" {SpecialToken.TAB_SYMBOL.value} ")
        token = re.sub(" +", " ", token)
        if use_bleu_tokenization:
            token = self.tokenizer_v14_international(token)
        token = re.sub(" +", " ", token)

        # split string prefix if one stands at the beginning
        regex = r"""([bruf]*) ((\"""|'''|"|')(?:(?!\3)(?:\\.|[^\\]))*\3)"""
        token = re.sub(regex, r"\1\2", token)

        for special_token, char in self.token2char.items():
            token = token.replace(special_token, char)

        token = token.replace("\r", "")

        return token


class PyTokenizer(Tokenizer):
    SPECIAL_TOKENS: List[str] = [token.value for token in SpecialToken]

    def __init__(self):
        super().__init__(
            special_tokens=self.SPECIAL_TOKENS,
            char2token=PYTHON_CHAR2TOKEN,
            token2char=PYTHON_TOKEN2CHAR,
        )

    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    def tokenize(
        self,
        text: str,
        keep_comments: bool = False,
        add_token_space: bool = True,
    ) -> str:
        # pylint: disable=too-many-nested-blocks
        try:
            assert isinstance(text, str)
            text = text.replace(r"\r", "")
            tokens = []

            try:
                text_bytes = BytesIO(text.encode("utf-8"))
                iterator = tokenize.tokenize(text_bytes.readline)
            except SyntaxError as err:
                raise err

            removed_docstr = 0
            while True:
                try:
                    token_info = next(iterator)
                except (
                    tokenize.TokenError,
                    IndentationError,
                    SyntaxError,
                    UnicodeDecodeError,
                ) as err:
                    raise Exception(
                        f'Impossible to parse tokens because icorrect source code "{text[0:30]}"...'
                    ) from err
                except StopIteration as err:
                    raise Exception("End of iterator before ENDMARKER token.") from err
                # pylint: disable=no-else-continue
                if token_info.type in [tokenize.ENCODING, tokenize.NL]:
                    continue
                elif token_info.type == tokenize.NEWLINE:
                    if removed_docstr == 1:
                        removed_docstr = 0
                        continue
                    tokens.append(SpecialToken.NEWLINE.value)
                elif token_info.type == tokenize.COMMENT:
                    if keep_comments:
                        com = self.process_string(
                            token_info.string,
                            is_comment=True,
                        )
                        if len(com) > 0:
                            tokens.append(com)
                    else:
                        continue
                elif token_info.type == tokenize.STRING:
                    if token_info.string == token_info.line.strip():  # docstring
                        if not keep_comments:
                            removed_docstr = 1
                            continue
                        else:
                            coms = self.process_string(
                                token_info.string,
                                is_comment=True,
                            )
                            if len(coms) > 0:
                                tokens.append(coms)
                            else:
                                removed_docstr = 1
                    else:
                        tokens.append(
                            self.process_string(
                                token_info.string,
                                is_comment=False,
                            )
                        )
                elif token_info.type == tokenize.INDENT:
                    tokens.append(SpecialToken.INDENT.value)
                elif token_info.type == tokenize.DEDENT:
                    # empty block
                    if tokens[-1] == SpecialToken.INDENT.value:
                        tokens = tokens[:-1]
                    else:
                        tokens.append(SpecialToken.DEDENT.value)
                elif token_info.type == tokenize.ENDMARKER:
                    tokens.append(SpecialToken.END_MARKER.value)
                    break
                else:
                    tokens.append(token_info.string)
            assert tokens[-1] == SpecialToken.END_MARKER.value, "Error, no end marker"
        except KeyboardInterrupt as err:
            raise err
        except BaseException:  # pylint: disable=broad-except
            tokens = []
        return self.stringify(tokens[:-1], add_token_space)

    def detokenize(
        self,
        tokens: Union[List[str], str],
        add_token_space: bool = True,
    ) -> str:
        try:
            assert isinstance(tokens, (str, list))
            if isinstance(tokens, list):
                tokens = " ".join(tokens)

            # restore missing spaces around special tokens
            tokens = self.destringify(tokens, add_token_space)

            # prepare tokenized string for splitting
            tokens = (
                tokens.replace(
                    SpecialToken.END_COM.value,
                    SpecialToken.NEWLINE.value,
                )
                .replace(
                    SpecialToken.SPACE.value,
                    SpecialToken.SPACE_TOKEN.value,
                )
                .replace(
                    SpecialToken.STR_NEWLINE.value,
                    SpecialToken.NEWLINE.value,
                )
            )

            # prepare tokenized lines for detokenization
            lines = tokens.split(SpecialToken.NEWLINE.value)

            tabs = ""
            for idx, line in enumerate(lines):
                line = line.strip()
                if line.startswith(f"{SpecialToken.INDENT.value} "):
                    tabs += "    "
                    line = line.replace(f"{SpecialToken.INDENT.value} ", tabs)
                elif line.startswith(SpecialToken.DEDENT.value):
                    number_dedent = line.count(SpecialToken.DEDENT.value)
                    tabs = tabs[4 * number_dedent :]
                    line = line.replace(SpecialToken.DEDENT.value, "")
                    line = line.strip()
                    line = tabs + line
                elif line == SpecialToken.DEDENT.value:
                    line = ""
                else:
                    line = tabs + line

                # python tokenizer ignores comment indentation
                # this way we check if comment should get more tabs
                try:
                    next_line = lines[idx + 1].strip()
                    if line.strip().startswith("#"):
                        if next_line.startswith(f"{SpecialToken.INDENT.value} "):
                            line = "    " + line
                except IndexError:
                    pass

                lines[idx] = line

            untokenized = "\n".join(lines)

            # detokenize string and comment
            untokenized = (
                untokenized.replace(SpecialToken.STR_NEWLINE.value, "\n")
                .replace(SpecialToken.TAB_SYMBOL.value, "\t")
                .replace(f" {SpecialToken.SPACE_TOKEN.value} ", " ")
                .replace(SpecialToken.SPACE_TOKEN.value, " ")
            )

            # detokenize imports
            untokenized = (
                untokenized.replace(". ", ".")
                .replace(" .", ".")
                .replace("import.", "import .")
                .replace("from.", "from .")
            )

            untokenized = untokenized.replace("> >", ">>").replace("< <", "<<")
            return untokenized
        except KeyboardInterrupt as err:
            raise err
        except BaseException:  # pylint: disable=broad-except
            return ""



from .configs import (
    DatasetConfig,
    HFDatasetConfig,
    LocalDatasetConfig,
    MultitaskDatasetConfig,
    PreTrainedTokenizerConfig,
    TrainingConfig,
)
from .data_module import DataModuleConfig, T5DataModule, T5DataModuleConfig
from .logger import LoggerConfig, WandbLogger, WandbLoggerConfig
from .model import ModelConfig, T5ModelForPreTraining, T5ModelForPreTrainingConfig
from .optimizers import (
    AdafactorConfig,
    AdagradConfig,
    AdamConfig,
    AdamWConfig,
    AutoOptimizer,
    OptimizerConfig,
)
from .schedulers import (
    AutoScheduler,
    ConstantSchedulerConfig,
    InverseSquareRootSchedulerConfig,
    LinearSchedulerConfig,
    PolynomialSchedulerConfig,
    SchedulerConfig,
)
from .tokenizer_model import (
    SentencePieceTokenizer,
    SentencePieceTokenizerConfig,
    TokenizerConfig,
)
from .tokenizer_trainer import (
    SentencePieceTrainer,
    SentencePieceTrainerConfig,
    TokenizerTrainerConfig,
)
from .trainer import T5Trainer, T5TrainerConfig, TrainerConfig



from dataclasses import dataclass
from typing import Optional

from omegaconf import MISSING


@dataclass
class TrainingConfig:
    output_dir: str = MISSING
    checkpoint_dir: Optional[str] = None
    cache_dir: Optional[str] = None
    overwrite_cache: bool = False
    push_to_hub: bool = False
    push_to_hub_model_id: Optional[str] = None
    push_to_hub_organization: Optional[str] = None
    push_to_hub_token: Optional[str] = None
    seed: int = 42



from dataclasses import dataclass, field
from typing import Any, Dict, Optional

import hydra
from omegaconf import MISSING


@dataclass
class DatasetConfig:
    _target_: str = "git_t5.DatasetConfig"
    column_name: Optional[str] = None


@dataclass
class LocalDatasetConfig(DatasetConfig):
    _target_: str = "git_t5.LocalDatasetConfig"
    dataset_path: str = MISSING


@dataclass
class HFDatasetConfig(DatasetConfig):
    _target_: str = "git_t5.HFDatasetConfig"
    dataset_name: str = MISSING
    dataset_config: Optional[str] = None


@dataclass
class MultitaskDatasetConfig(DatasetConfig):
    _target_: str = "git_t5.MultitaskDatasetConfig"
    tasks: Dict[str, Any] = field(default_factory=dict)

    def resolve(self) -> None:
        """Resolves dynamic task configs into instantiated instances."""
        for task_name, task_config in self.tasks.items():
            self.tasks[task_name] = hydra.utils.instantiate(task_config)



from .dataset import (
    DatasetConfig,
    HFDatasetConfig,
    LocalDatasetConfig,
    MultitaskDatasetConfig,
)
from .tokenizer import PreTrainedTokenizerConfig
from .training import TrainingConfig



import itertools
import math
import typing
from abc import abstractmethod
from itertools import zip_longest
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Union

import datasets as hfds
import jax
import more_itertools
import numpy as np
from flax.training.common_utils import shard
from torch.utils import data

T = typing.TypeVar("T")


def flatten(nested_dict: Dict[str, List[Any]]) -> List[Dict[str, Any]]:
    result = []
    packed_values = zip_longest(*nested_dict.values())
    for packed_value in packed_values:
        flat_dict = {}
        for key, value in zip(nested_dict.keys(), packed_value):
            if value is None:
                continue
            flat_dict[key] = value
        result.append(flat_dict)
    return result


def chunked_iterator(
    iterable: Iterable[T], chunk_size: int, drop_last: bool
) -> Iterator[List[T]]:
    for chunk in more_itertools.chunked(iterable, chunk_size):
        if len(chunk) != chunk_size and drop_last:
            continue
        yield chunk


class Dataset(data.IterableDataset):
    def __getitem__(self, index: int) -> Dict[str, np.ndarray]:
        raise RuntimeError("`Dataset` subclasses do not support indexing.")

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def __iter__(self) -> Iterator[Dict[str, np.ndarray]]:
        raise NotImplementedError


class T5Dataset(Dataset):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        dataset: hfds.Dataset,
        batch_size: int,
        collate_fn: Callable[..., Any],
        shuffle: bool,
        drop_last: bool = False,
        seed: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.dataset = dataset
        self.batch_size = batch_size
        self.collate_fn = collate_fn
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.seed = seed

    def __len__(self) -> int:
        instance_count = len(self.dataset)
        if self.drop_last or instance_count % self.batch_size == 0:
            return instance_count // self.batch_size
        return 1 + (instance_count // self.batch_size)

    def __iter__(self) -> Iterator[Dict[str, np.ndarray]]:
        if self.shuffle:
            self.dataset = self.dataset.shuffle(self.seed)

        # get the total number of instances per epoch
        length = len(self.dataset)

        # calculate the number of batches per worker
        worker_info = data.get_worker_info()
        if worker_info is not None:
            num_workers = float(worker_info.num_workers) * self.batch_size
            per_worker = int(math.ceil(length / num_workers))
            iter_start = worker_info.id * per_worker
            iter_stop = min(iter_start + per_worker, length)
        else:
            iter_start = 0
            iter_stop = length

        # split the batched instances across the workers
        batches = chunked_iterator(self.dataset, self.batch_size, self.drop_last)
        batches = itertools.islice(batches, iter_start, iter_stop)
        for batch in batches:
            batch = self.collate_fn(batch)
            batch = jax.tree_map(shard, batch)
            yield batch


class T5MultitaskDataset(Dataset):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        datasets: Union[Dict[str, hfds.Dataset], List[hfds.Dataset]],
        batch_size: int,
        collate_fn: Callable[..., Any],
        shuffle: bool,
        drop_last: bool = True,
        seed: Optional[int] = None,
        limit_instances: Optional[int] = None,
    ) -> None:
        super().__init__()

        if isinstance(datasets, list):
            datasets = {
                f"dataset_{idx}": dataset for idx, dataset in enumerate(datasets)
            }

        self.datasets = datasets
        self.batch_size = batch_size
        self.collate_fn = collate_fn
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.seed = seed
        self.limit_instances = limit_instances

    def __len__(self) -> int:
        dataset_counts: Dict[str, int]
        if self.limit_instances is None:
            dataset_counts = {
                name: len(dataset) for name, dataset in self.datasets.items()
            }
        else:
            dataset_counts = self.get_task_num_instances(
                self.datasets,
                self.limit_instances,
            )

        return self.count_batches(dataset_counts)

    def __iter__(self) -> Iterator[Dict[str, np.ndarray]]:
        if self.shuffle:
            self.datasets = {
                name: dataset.shuffle(self.seed)
                for name, dataset in self.datasets.items()
            }

        # get the joined instances across all datasets
        epoch_instances = self.get_epoch_instances()
        epoch_instances = more_itertools.roundrobin(*epoch_instances.values())

        # get the total number of instances per epoch
        epoch_counts = self.get_dataset_counts()
        length = sum(epoch_counts.values())

        # calculate the number of batches per worker
        worker_info = data.get_worker_info()
        if worker_info is not None:
            num_workers = float(worker_info.num_workers) * self.batch_size
            per_worker = int(math.ceil(length / num_workers))
            iter_start = worker_info.id * per_worker
            iter_stop = min(iter_start + per_worker, length)
        else:
            iter_start = 0
            iter_stop = length

        # split the batched instances across the workers
        batches = chunked_iterator(epoch_instances, self.batch_size, self.drop_last)
        batches = itertools.islice(batches, iter_start, iter_stop)
        for batch in batches:
            batch = self.collate_fn(batch)
            batch = jax.tree_map(shard, batch)
            yield batch

    def get_epoch_instances(self) -> Dict[str, Iterable[Dict[str, np.ndarray]]]:
        dataset_counts = self.get_dataset_counts()
        return {
            key: itertools.islice(self.datasets[key], count)
            for key, count in dataset_counts.items()
        }

    def get_dataset_counts(self) -> Dict[str, int]:
        dataset_counts: Dict[str, int]
        if self.limit_instances is None:
            dataset_counts = {
                name: len(dataset) for name, dataset in self.datasets.items()
            }
        else:
            dataset_counts = self.get_task_num_instances(
                self.datasets,
                self.limit_instances,
            )
        return dataset_counts

    def get_task_num_instances(
        self, datasets: Dict[str, hfds.Dataset], limit_instances: int
    ) -> Dict[str, int]:
        dataset_proportions = self.get_task_proportions(datasets)
        total_proportions = sum(dataset_proportions.values())
        return {
            key: math.floor(proportion * limit_instances / total_proportions)
            for key, proportion in dataset_proportions.items()
        }

    def get_task_proportions(
        self, datasets: Dict[str, hfds.Dataset]
    ) -> Dict[str, float]:
        return {key: 1 / len(datasets) for key in datasets}

    def count_batches(self, dataset_counts: Dict[str, int]) -> int:
        instance_count = sum(dataset_counts.values())
        if self.drop_last or instance_count % self.batch_size == 0:
            return instance_count // self.batch_size
        return 1 + (instance_count // self.batch_size)



from functools import wraps
from typing import Callable, TypeVar

import jax

T = TypeVar("T", bound=Callable)


def rank_zero_only(fn: T) -> T:
    @wraps(fn)
    def wrapped_fn(*args, **kwargs):
        if jax.process_index() == 0:
            return fn(*args, **kwargs)

    return wrapped_fn  # type: ignore



from .code_tokenizer import PyTokenizer, Tokenizer
from .common import stack_mappings
from .distributed import rank_zero_only
from .html_stripper import HTMLStripper
from .resolvers import resolve_object
from .unidiff_builder import UnidiffBuilder



from dataclasses import dataclass
from typing import Dict, List

import numpy as np
from transformers import PreTrainedTokenizerBase

from .data_preprocessing import (
    noise_span_to_unique_sentinel,
    nonnoise_span_to_unique_sentinel,
)


def shift_tokens_right(
    input_ids: np.ndarray,
    pad_token_id: int,
    decoder_start_token_id: int,
) -> np.ndarray:
    """
    Shift input ids one token to the right.
    """
    shifted_input_ids = np.zeros_like(input_ids)
    shifted_input_ids[..., 1:] = input_ids[..., :-1]
    shifted_input_ids[..., 0] = decoder_start_token_id

    # replace possible -100 values in labels by `pad_token_id`
    shifted_input_ids = np.where(
        shifted_input_ids == -100,
        pad_token_id,
        shifted_input_ids,
    )

    return shifted_input_ids


@dataclass
class DataCollatorForT5MLM:
    """
    Data collator used for T5 span-masked language modeling.
    It is made sure that after masking the inputs are of length `data_args.max_seq_length` and targets are also of fixed length.
    For more information on how T5 span-masked language modeling works, one can take a look
    at the `official paper <https://arxiv.org/pdf/1910.10683.pdf>`__
    or the `official code for preprocessing <https://github.com/google-research/text-to-text-transfer-transformer/blob/master/t5/data/preprocessors.py>`__ .
    Args:
        tokenizer (:class:`~transformers.PreTrainedTokenizer` or :class:`~transformers.PreTrainedTokenizerFast`):
            The tokenizer used for encoding the data.
        noise_density (:obj:`float`):
            The probability with which to (randomly) mask tokens in the input.
        mean_noise_span_length (:obj:`float`):
            The average span length of the masked tokens.
        input_length (:obj:`int`):
            The expected input length after masking.
        target_length (:obj:`int`):
            The expected target length after masking.
        eos_token_id: (:obj:`int`):
            The eos token id of the model.
        pad_token_id: (:obj:`int`):
            The pad token id of the model.
        decoder_start_token_id: (:obj:`int):
            The decoder start token id of the model.
        sentinel_token_id: (:obj:`int):
            The first sentinel token id of the model.
        sentinel_tokens_reversed: (:obj:`bool):
            Whether sentinel tokens arranged in reverse order.
    """

    tokenizer: PreTrainedTokenizerBase
    noise_density: float
    mean_noise_span_length: float
    input_length: int
    target_length: int
    eos_token_id: int
    pad_token_id: int
    decoder_start_token_id: int
    sentinel_token_id: int
    sentinel_tokens_reversed: bool = False

    def __call__(self, examples: List[Dict[str, np.ndarray]]) -> Dict[str, np.ndarray]:
        # add span corruption noise for each example
        examples = [self.denoise(example) for example in examples]

        # collate input examples
        batch = {
            column: np.array([examples[idx][column] for idx in range(len(examples))])
            for column in examples[0].keys()
        }

        if batch["input_ids"].shape[-1] != self.input_length:
            raise ValueError(
                f"`input_ids` are incorrectly preprocessed. `input_ids` length is {batch['input_ids'].shape[-1]}, but should be {self.input_length}."
            )

        if batch["labels"].shape[-1] != self.target_length:
            raise ValueError(
                f"`labels` are incorrectly preprocessed. `labels` length is {batch['labels'].shape[-1]}, but should be {self.target_length}."
            )

        return batch

    def denoise(self, features: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        tokens = features["input_ids"]
        noise_mask = self.random_spans_noise_mask(np.size(tokens))

        input_ids = self.append_eos_token(
            noise_span_to_unique_sentinel(
                tokens,
                noise_mask,
                self.sentinel_token_id,
                self.sentinel_tokens_reversed,
            )
        )
        labels = self.append_eos_token(
            nonnoise_span_to_unique_sentinel(
                tokens,
                noise_mask,
                self.sentinel_token_id,
                self.sentinel_tokens_reversed,
            )
        )
        decoder_input_ids = shift_tokens_right(
            labels,
            self.pad_token_id,
            self.decoder_start_token_id,
        )

        return {
            "input_ids": input_ids,
            "labels": labels,
            "decoder_input_ids": decoder_input_ids,
        }

    def append_eos_token(self, tokens: np.ndarray) -> np.ndarray:
        return np.concatenate((tokens, [self.eos_token_id]), axis=0)

    def random_spans_noise_mask(self, length: int) -> np.ndarray:
        """This function is copy of `random_spans_helper <https://github.com/google-research/text-to-text-transfer-transformer/blob/84f8bcc14b5f2c03de51bd3587609ba8f6bbd1cd/t5/data/preprocessors.py#L2682>`__ .
        Noise mask consisting of random spans of noise tokens.
        The number of noise tokens and the number of noise spans and non-noise spans
        are determined deterministically as follows:
        num_noise_tokens = round(length * noise_density)
        num_nonnoise_spans = num_noise_spans = round(num_noise_tokens / mean_noise_span_length)
        Spans alternate between non-noise and noise, beginning with non-noise.
        Subject to the above restrictions, all masks are equally likely.
        Args:
            length: an int32 scalar (length of the incoming token sequence)
            noise_density: a float - approximate density of output mask
            mean_noise_span_length: a number
        Returns:
            a boolean tensor with shape [length]
        """

        orig_length = length

        num_noise_tokens = int(np.round(length * self.noise_density))
        # avoid degeneracy by ensuring positive numbers of noise and nonnoise tokens.
        num_noise_tokens = min(max(num_noise_tokens, 1), length - 1)
        num_noise_spans = int(np.round(num_noise_tokens / self.mean_noise_span_length))

        # avoid degeneracy by ensuring positive number of noise spans
        num_noise_spans = max(num_noise_spans, 1)
        num_nonnoise_tokens = length - num_noise_tokens

        # pick the lengths of the noise spans and the non-noise spans
        def _random_segmentation(num_items: int, num_segments: int) -> np.ndarray:
            """Partition a sequence of items randomly into non-empty segments.
            Args:
                num_items: an integer scalar > 0
                num_segments: an integer scalar in [1, num_items]
            Returns:
                a Tensor with shape [num_segments] containing positive integers that add
                up to num_items
            """
            mask_indices = np.arange(num_items - 1) < (num_segments - 1)
            np.random.shuffle(mask_indices)
            first_in_segment = np.pad(mask_indices, [[1, 0]])
            segment_id = np.cumsum(first_in_segment)
            _, segment_length = np.unique(segment_id, return_counts=True)
            return segment_length

        noise_span_lengths = _random_segmentation(num_noise_tokens, num_noise_spans)
        nonnoise_span_lengths = _random_segmentation(
            num_nonnoise_tokens, num_noise_spans
        )

        interleaved_span_lengths = np.reshape(
            np.stack([nonnoise_span_lengths, noise_span_lengths], axis=1),
            [num_noise_spans * 2],
        )
        span_starts = np.cumsum(interleaved_span_lengths)[:-1]
        span_start_indicator = np.zeros((length,), dtype=np.int8)
        span_start_indicator[span_starts] = True
        span_num = np.cumsum(span_start_indicator)
        is_noise = np.equal(span_num % 2, 1)

        return is_noise[:orig_length]



from git_t5.core import (
    T5DataModule,
    T5DataModuleConfig,
    T5ModelForPreTraining,
    T5ModelForPreTrainingConfig,
    T5Trainer,
    T5TrainerConfig,
    TrainerConfig,
)
from git_t5.core.configs import (
    DatasetConfig,
    HFDatasetConfig,
    LocalDatasetConfig,
    MultitaskDatasetConfig,
    PreTrainedTokenizerConfig,
    TrainingConfig,
)



import os
from dataclasses import dataclass

import hydra
from git_t5.core import (
    T5DataModule,
    T5DataModuleConfig,
    T5ModelForPreTraining,
    T5ModelForPreTrainingConfig,
    T5Trainer,
    T5TrainerConfig,
    WandbLogger,
    WandbLoggerConfig,
)
from git_t5.core.configs import (
    HFDatasetConfig,
    LocalDatasetConfig,
    MultitaskDatasetConfig,
)
from git_t5.core.optimizers import (
    AdafactorConfig,
    AdagradConfig,
    AdamConfig,
    AdamWConfig,
    OptimizerConfig,
)
from git_t5.core.schedulers import (
    ConstantSchedulerConfig,
    InverseSquareRootSchedulerConfig,
    LinearSchedulerConfig,
    PolynomialSchedulerConfig,
)
from hydra.core.config_store import ConfigStore

from .config import DefaultConfig, register_base_configs


@dataclass
class Config(DefaultConfig):
    data: T5DataModuleConfig = T5DataModuleConfig()
    model: T5ModelForPreTrainingConfig = T5ModelForPreTrainingConfig()
    optimizer: OptimizerConfig = OptimizerConfig()
    trainer: T5TrainerConfig = T5TrainerConfig()
    logger: WandbLoggerConfig = WandbLoggerConfig()


def register_optimizers(cs: ConfigStore) -> None:
    cs.store(
        group="optimizer",
        name="base_adam",
        node=AdamConfig,
    )
    cs.store(
        group="optimizer",
        name="base_adamw",
        node=AdamWConfig,
    )
    cs.store(
        group="optimizer",
        name="base_adafactor",
        node=AdafactorConfig,
    )
    cs.store(
        group="optimizer",
        name="base_adagrad",
        node=AdagradConfig,
    )


def register_schedulers(cs: ConfigStore) -> None:
    cs.store(
        group="optimizer/scheduler",
        name="base_polynomial",
        node=PolynomialSchedulerConfig,
    )
    cs.store(
        group="optimizer/scheduler",
        name="base_inverse_square_root",
        node=InverseSquareRootSchedulerConfig,
    )
    cs.store(
        group="optimizer/scheduler",
        name="base_linear",
        node=LinearSchedulerConfig,
    )
    cs.store(
        group="optimizer/scheduler",
        name="base_constant",
        node=ConstantSchedulerConfig,
    )


def register_datasets(cs: ConfigStore) -> None:
    cs.store(
        group="dataset",
        name="base_huggingface_dataset",
        node=HFDatasetConfig,
    )
    cs.store(
        group="dataset",
        name="base_local_dataset",
        node=LocalDatasetConfig,
    )
    cs.store(
        group="dataset",
        name="base_multitask_dataset",
        node=MultitaskDatasetConfig,
    )


def register_configs() -> None:
    cs = ConfigStore.instance()
    cs.store(name="default", node=Config)
    cs.store(
        group="data",
        name="base_data",
        node=T5DataModuleConfig,
    )
    cs.store(
        group="model",
        name="base_model",
        node=T5ModelForPreTrainingConfig,
    )
    cs.store(
        group="trainer",
        name="base_trainer",
        node=T5TrainerConfig,
    )
    cs.store(
        group="logger",
        name="base_logger",
        node=WandbLoggerConfig,
    )

    register_optimizers(cs)
    register_schedulers(cs)
    register_datasets(cs)


@hydra.main(config_path="conf", config_name="config_model")
def hydra_entry(cfg: Config) -> None:
    # disable rust iterators multithreading
    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    logger = WandbLogger(cfg.logger)
    model = T5ModelForPreTraining.from_config(cfg)
    data_module = T5DataModule.from_config(cfg)

    trainer = T5Trainer(
        config=cfg,
        model=model,
        data_module=data_module,
        logger=logger,
    )

    trainer.fit()


def main() -> None:
    register_base_configs()
    register_configs()
    hydra_entry()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Optional

import jax.numpy as jnp
import optax
from git_t5.utils import resolve_object

from .schedulers import AutoScheduler, SchedulerConfig


@dataclass
class OptimizerConfig:
    scheduler: SchedulerConfig = SchedulerConfig()


@dataclass
class AdamConfig(OptimizerConfig):
    b1: float = 0.9
    b2: float = 0.999
    eps: float = 1e-8
    eps_root: float = 0.001


@dataclass
class AdamWConfig(OptimizerConfig):
    b1: float = 0.9
    b2: float = 0.999
    eps: float = 1e-8
    eps_root: float = 0.001
    weight_decay: float = 1e-4


@dataclass
class AdafactorConfig(OptimizerConfig):
    min_dim_size_to_factor: int = 128
    decay_rate: float = 0.8
    decay_offset: int = 0
    multiply_by_parameter_scale: bool = True
    clipping_threshold: Optional[float] = 1.0
    momentum: Optional[float] = None
    weight_decay_rate: Optional[float] = None
    dtype_momentum: str = "float32"
    eps: float = 1e-30
    factored: bool = True


@dataclass
class AdagradConfig(OptimizerConfig):
    initial_accumulator_value: float = 0.1
    eps: float = 1e-7


def adam(config: AdamConfig) -> optax.GradientTransformation:
    scheduler = AutoScheduler.from_config(config.scheduler)
    return optax.adam(
        learning_rate=scheduler,
        b1=config.b1,
        b2=config.b2,
        eps=config.eps,
        eps_root=config.eps_root,
    )


def adamw(config: AdamWConfig) -> optax.GradientTransformation:
    scheduler = AutoScheduler.from_config(config.scheduler)
    return optax.adamw(
        learning_rate=scheduler,
        b1=config.b1,
        b2=config.b2,
        eps=config.eps,
        eps_root=config.eps_root,
        weight_decay=config.weight_decay,
    )


def adafactor(config: AdafactorConfig) -> optax.GradientTransformation:
    scheduler = AutoScheduler.from_config(config.scheduler)
    return optax.adafactor(
        learning_rate=scheduler,
        min_dim_size_to_factor=config.min_dim_size_to_factor,
        decay_rate=config.decay_rate,
        decay_offset=config.decay_offset,
        multiply_by_parameter_scale=config.multiply_by_parameter_scale,  # type: ignore
        clipping_threshold=config.clipping_threshold,
        momentum=config.momentum,
        dtype_momentum=getattr(jnp, config.dtype_momentum),
        weight_decay_rate=config.weight_decay_rate,
        eps=config.eps,
        factored=config.factored,
    )


def adagrad(config: AdagradConfig) -> optax.GradientTransformation:
    scheduler = AutoScheduler.from_config(config.scheduler)
    return optax.adagrad(
        learning_rate=scheduler,
        initial_accumulator_value=config.initial_accumulator_value,
        eps=config.eps,
    )


class AutoOptimizer:
    @staticmethod
    def from_config(config: OptimizerConfig) -> optax.GradientTransformation:
        config = resolve_object(config)
        optimizer: optax.GradientTransformation
        if isinstance(config, AdamConfig):
            optimizer = adam(config)
        elif isinstance(config, AdamWConfig):
            optimizer = adamw(config)
        elif isinstance(config, AdafactorConfig):
            optimizer = adafactor(config)
        elif isinstance(config, AdagradConfig):
            optimizer = adagrad(config)
        else:
            raise ValueError(f"Unknown optimizer type: {type(config)}.")
        return optimizer



from html.parser import HTMLParser
from io import StringIO


class HTMLStripper(HTMLParser):
    def __init__(self) -> None:
        super().__init__()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def reset(self) -> None:
        super().reset()
        self.text = StringIO()

    def strip_tags(self, text: str) -> str:
        self.reset()
        self.feed(text)
        return self.get_data()

    def handle_data(self, data: str) -> None:
        self.text.write(data)

    def error(self, message: str) -> None:
        raise RuntimeError(message)

    def get_data(self) -> str:
        return self.text.getvalue()



import os
import shutil
from dataclasses import dataclass
from typing import Iterator, List, Optional, Union

import datasets
from omegaconf import MISSING
from tokenizers.implementations.base_tokenizer import BaseTokenizer
from transformers import T5TokenizerFast

from .tokenizer_model import (
    SentencePieceTokenizer,
    SentencePieceTokenizerConfig,
    TokenizerConfig,
)


def batch_iterator(
    dataset: datasets.Dataset,
    column_name: str,
    batch_size: int = 1024,
) -> Iterator[Iterator[str]]:
    for i in range(0, len(dataset), batch_size):
        yield dataset[i : i + batch_size][column_name]  # type: ignore


def save(tokenizer: BaseTokenizer, output_path: Union[str, os.PathLike]) -> str:
    # remove previous dir if one exists
    if os.path.exists(output_path):
        shutil.rmtree(output_path)
    # make a new output directory
    os.makedirs(output_path, exist_ok=True)
    # save pretrained tokenizer file
    tokenizer_file = os.path.join(output_path, "tokenizer.json")
    tokenizer.save(tokenizer_file)
    return tokenizer_file


@dataclass
class TokenizerTrainerConfig:
    save_directory: str = MISSING
    tokenizer: TokenizerConfig = TokenizerConfig()


@dataclass
class SentencePieceTrainerConfig(TokenizerTrainerConfig):
    dataset_column: str = MISSING
    dataset_path: Optional[str] = None
    dataset_files: Optional[List[str]] = None
    tokenizer: SentencePieceTokenizerConfig = SentencePieceTokenizerConfig()


@dataclass
class SentencePieceTrainer:
    config: SentencePieceTrainerConfig
    tokenizer: SentencePieceTokenizer

    @classmethod
    def from_config(cls, config: SentencePieceTrainerConfig) -> "SentencePieceTrainer":
        tokenizer = SentencePieceTokenizer(config.tokenizer)
        return SentencePieceTrainer(config, tokenizer)

    def train(self) -> None:
        if self.config.dataset_path:
            dataset = datasets.load_from_disk(self.config.dataset_path)
            if isinstance(dataset, datasets.DatasetDict):
                dataset = dataset["train"]

            dataset_iter = batch_iterator(dataset, self.config.dataset_column)
            self.tokenizer.train_from_iterator(dataset_iter)
        elif self.config.dataset_files:
            self.tokenizer.train_from_files(self.config.dataset_files)
        else:
            raise ValueError("`dataset_path` or `dataset_files` must be specified.")

        # save pre-trained rust tokenizer
        tokenizer_file = save(self.tokenizer, self.config.save_directory)

        # create a fast tokenizer wrapper around rust backend
        tokenizer_fast = T5TokenizerFast(
            vocab_file=None,
            tokenizer_file=tokenizer_file,
            model_max_length=self.tokenizer.config.model_max_length,
        )

        # concatenate extra_ids and user defined additional_special_tokens
        special_tokens = self.tokenizer.config.additional_special_tokens
        additional_special_tokens = tokenizer_fast.additional_special_tokens
        additional_special_tokens += list(special_tokens)

        tokenizer_fast.add_special_tokens(
            {"additional_special_tokens": additional_special_tokens}  # type: ignore
        )

        # save t5 fast tokenizer in a compact format
        tokenizer_fast.save_pretrained(
            save_directory=self.config.save_directory,
            legacy_format=False,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['git_t5',
 'git_t5.cli',
 'git_t5.cli.conf',
 'git_t5.core',
 'git_t5.core.configs',
 'git_t5.data',
 'git_t5.tests',
 'git_t5.utils']

package_data = \
{'': ['*'],
 'git_t5.cli.conf': ['data/*',
                     'dataset/*',
                     'logger/*',
                     'model/*',
                     'optimizer/*',
                     'optimizer/scheduler/*',
                     'tokenizer/*',
                     'tokenizer_trainer/*',
                     'tokenizer_trainer/tokenizer/*',
                     'trainer/*',
                     'training/*']}

install_requires = \
['datasets>=1.11.0,<2.0.0',
 'flax>=0.3.4,<0.4.0',
 'hydra-core>=1.1.1,<2.0.0',
 'jax>=0.2.19,<0.3.0',
 'more-itertools>=8.8.0,<9.0.0',
 'optax>=0.0.9,<0.0.10',
 'sacrebleu>=1.5.1,<2.0.0',
 'torch>=1.9.0,<2.0.0',
 'transformers>=4.10.0,<5.0.0',
 'wandb>=0.12.1,<0.13.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.10.0.2,<4.0.0.0']}

entry_points = \
{'console_scripts': ['gt5-train-model = git_t5.cli.train_model:main',
                     'gt5-train-tokenizer = git_t5.cli.train_tokenizer:main']}

setup_kwargs = {
    'name': 'git-t5',
    'version': '0.2.0',
    'description': 'Open source machine learning framework for training T5 models on source code in JAX/Flax.',
    'long_description': '# git-t5',
    'author': 'mozharovsky',
    'author_email': 'mozharovsky@live.com',
    'maintainer': 'mozharovsky',
    'maintainer_email': 'mozharovsky@live.com',
    'url': 'https://github.com/formermagic/git-t5',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

import jax
import jax.numpy as jnp
import optax
from flax import traverse_util
from flax.training.common_utils import onehot
from flax.training.train_state import TrainState
from transformers import (
    CONFIG_MAPPING,
    AutoTokenizer,
    FlaxT5ForConditionalGeneration,
    PreTrainedTokenizerBase,
    T5Config,
)

if TYPE_CHECKING:
    from git_t5.cli.train_model import Config
    from git_t5.core.trainer import T5Trainer
else:
    Config = Any
    T5Trainer = Any

MAX_VALUE = (2 << 30) - 1


def decay_mask_fn(params: Dict[str, Any]) -> Dict[str, Any]:
    """
    We use Optax's "masking" functionality to not apply weight decay
    to bias and LayerNorm scale parameters. decay_mask_fn returns a
    mask boolean with the same structure as the parameters.
    The mask is True for parameters that should be decayed.
    """

    flat_params = traverse_util.flatten_dict(params)
    flat_mask = {
        path: (
            path[-1] != "bias"  # type: ignore
            and path[-2:]
            not in [("layer_norm", "scale"), ("final_layer_norm", "scale")]
        )
        for path in flat_params
    }

    return traverse_util.unflatten_dict(flat_mask)


@dataclass
class ModelConfig:
    pass


@dataclass
class T5ModelForPreTrainingConfig(ModelConfig):
    model_path: Optional[str] = None
    model_type: Optional[str] = None
    config_name: Optional[str] = None
    dtype: str = "float32"


@dataclass
class T5ModelForPreTraining:
    config: Config
    model: FlaxT5ForConditionalGeneration
    tokenizer: PreTrainedTokenizerBase
    trainer: Optional[T5Trainer] = None

    @classmethod
    def from_config(cls, config: Config) -> "T5ModelForPreTraining":
        tokenizer = cls.load_tokenizer(config)
        model = cls.load_model(config, tokenizer)
        return T5ModelForPreTraining(config, model, tokenizer)

    @partial(
        jax.pmap,
        axis_name="batch",
        static_broadcasted_argnums=(0,),
        donate_argnums=(1,),
    )
    def train_step(
        self,
        state: TrainState,
        dropout_rng: jnp.ndarray,
        batch: Dict[str, Any],
    ) -> Tuple[TrainState, Dict[str, jnp.ndarray], jnp.ndarray]:
        def loss_fn(params: Dict[str, Any]) -> float:
            labels = batch.pop("labels")
            outputs = state.apply_fn(
                **batch,
                params=params,
                dropout_rng=dropout_rng,
                train=True,
            )

            # compute loss
            logits = outputs[0]
            labels = onehot(labels, logits.shape[-1])
            loss = optax.softmax_cross_entropy(logits, labels).mean()

            return loss

        dropout_rng, new_dropout_rng = jax.random.split(dropout_rng)  # type: ignore

        # compute loss & gradients
        grad_fn = jax.value_and_grad(loss_fn)
        loss, grads = grad_fn(state.params)
        # compute perplexity
        perplexity = jnp.exp(loss).clip(0, MAX_VALUE)
        # apply computed gradients
        grads = jax.lax.pmean(grads, "batch")
        state = state.apply_gradients(grads=grads)
        # prepare train metrics
        metrics = {"loss": loss, "perplexity": perplexity}
        metrics = jax.lax.pmean(metrics, axis_name="batch")

        return state, metrics, new_dropout_rng  # type: ignore

    @partial(
        jax.pmap,
        axis_name="batch",
        static_broadcasted_argnums=(0,),
        donate_argnums=(1,),
    )
    def valid_step(
        self,
        state: TrainState,
        batch: Dict[str, Any],
    ) -> Dict[str, jnp.ndarray]:
        labels = batch.pop("labels")
        outputs = state.apply_fn(**batch, params=state.params, train=False)

        # compute loss
        logits = outputs[0]
        labels_onehot = onehot(labels, logits.shape[-1])
        loss = optax.softmax_cross_entropy(logits, labels_onehot).mean()
        # compute perplexity
        perplexity = jnp.exp(loss).mean().clip(0, MAX_VALUE)
        # compute accuracy
        accuracy = jnp.equal(jnp.argmax(logits, axis=-1), labels).mean()
        # prepare validation metrics
        metrics = {"loss": loss, "perplexity": perplexity, "accuracy": accuracy}
        metrics = jax.lax.pmean(metrics, axis_name="batch")

        return metrics

    @classmethod
    def load_model(
        cls,
        config: Config,
        tokenizer: PreTrainedTokenizerBase,
    ) -> FlaxT5ForConditionalGeneration:
        if config.model.model_path is not None:
            model = FlaxT5ForConditionalGeneration.from_pretrained(
                config.model.model_path,
                config=config,
                seed=config.training.seed,
                dtype=getattr(jnp, config.model.dtype),
            )
        else:
            model = FlaxT5ForConditionalGeneration(
                cls.load_config(config, tokenizer),
                seed=config.training.seed,
                dtype=getattr(jnp, config.model.dtype),
            )

        return model

    @classmethod
    def load_config(
        cls,
        config: Config,
        tokenizer: PreTrainedTokenizerBase,
    ) -> T5Config:
        vocab_size = len(tokenizer)  # type: ignore
        if config.model.config_name is not None:
            model_config = T5Config.from_pretrained(
                config.model.config_name,
                cache_dir=config.training.cache_dir,
                vocab_size=vocab_size,
            )
        elif config.model.model_path is not None:
            model_config = T5Config.from_pretrained(
                config.model.model_path,
                cache_dir=config.training.cache_dir,
                vocab_size=vocab_size,
            )
        else:
            model_config = CONFIG_MAPPING[config.model_type]()  # type: ignore

        # make sure the type matches the config
        assert isinstance(model_config, T5Config)
        return model_config

    @classmethod
    def load_tokenizer(cls, config: Config) -> PreTrainedTokenizerBase:
        tokenizer = AutoTokenizer.from_pretrained(
            config.tokenizer.tokenizer_path,
            cache_dir=config.training.cache_dir,
            use_fast=config.tokenizer.use_fast,
        )

        return tokenizer



import difflib
import re
from typing import List

from .code_tokenizer import Tokenizer


class UnidiffBuilder:
    def __init__(self, tokenizer: Tokenizer) -> None:
        self.tokenizer = tokenizer

    def build(
        self,
        source_content: str,
        target_content: str,
        source_filename: str,
        target_filename: str,
        context_size: int = 3,
        lineterm: str = "",
    ) -> str:
        source_lines = self.tokenize(source_content)
        target_lines = self.tokenize(target_content)

        lines = difflib.unified_diff(
            source_lines,
            target_lines,
            source_filename,
            target_filename,
            n=context_size,
            lineterm=lineterm,
        )

        lines = [self.prepare_line(line) for line in lines]

        return " ".join(lines)

    def prepare_line(self, line: str) -> str:
        # remove common metadata string
        line = re.sub(r"@@ .* @@", "<newline>", line)
        # add a space to the succeeding token after `-` token
        line = re.sub(r"^\-([^-]+)", r"- \g<1>", line)
        # add a space to the succeeding token after `+` token
        line = re.sub(r"^\+([^+]+)", r"+ \g<1>", line)
        return line

    def tokenize(self, content: str) -> List[str]:
        content = self.tokenizer.tokenize(
            content,
            keep_comments=False,
            add_token_space=True,
        )
        content = content.replace("<newline>", "<newline>\n")
        content_lines = [line.strip() for line in content.split("\n")]
        return content_lines



from dataclasses import dataclass

from git_t5.core import (
    DataModuleConfig,
    DatasetConfig,
    LoggerConfig,
    ModelConfig,
    OptimizerConfig,
    PreTrainedTokenizerConfig,
    SchedulerConfig,
    TokenizerConfig,
    TokenizerTrainerConfig,
    TrainerConfig,
    TrainingConfig,
)
from hydra.core.config_store import ConfigStore


@dataclass
class DefaultConfig:
    tokenizer_trainer: TokenizerTrainerConfig = TokenizerTrainerConfig()
    tokenizer: PreTrainedTokenizerConfig = PreTrainedTokenizerConfig()
    dataset: DatasetConfig = DatasetConfig()
    model: ModelConfig = ModelConfig()
    data: DataModuleConfig = DataModuleConfig()
    optimizer: OptimizerConfig = OptimizerConfig()
    trainer: TrainerConfig = TrainerConfig()
    training: TrainingConfig = TrainingConfig()
    logger: LoggerConfig = LoggerConfig()


def register_base_configs() -> None:
    cs = ConfigStore.instance()
    cs.store(name="default", node=DefaultConfig)
    cs.store(
        group="tokenizer_trainer",
        name="base_tokenizer_trainer",
        node=TokenizerTrainerConfig,
    )
    cs.store(
        group="tokenizer_trainer/tokenizer",
        name="base_tokenizer",
        node=TokenizerConfig,
    )
    cs.store(
        group="tokenizer",
        name="base_pretrained_tokenizer",
        node=PreTrainedTokenizerConfig,
    )
    cs.store(
        group="dataset",
        name="base_dataset",
        node=DatasetConfig,
    )
    cs.store(
        group="model",
        name="base_model",
        node=ModelConfig,
    )
    cs.store(
        group="data",
        name="base_data",
        node=DataModuleConfig,
    )
    cs.store(
        group="optimizer",
        name="base_optimizer",
        node=OptimizerConfig,
    )
    cs.store(
        group="optimizer/scheduler",
        name="base_scheduler",
        node=SchedulerConfig,
    )
    cs.store(
        group="trainer",
        name="base_trainer",
        node=TrainerConfig,
    )
    cs.store(
        group="training",
        name="base_training",
        node=TrainingConfig,
    )
    cs.store(
        group="logger",
        name="base_logger",
        node=LoggerConfig,
    )
